package org.sc3d.apt.sss.v3;

/** Represents an escape sequence in a literal string or literal character. This is not a subclass of Token, because escape sequences are simply incorporated into the surrounding Token. This class exists only as a way of sharing the code that lexes escape sequences. */
public class Escape {
  /** Constructs an Escape given values for its fields. */
  public Escape(
    Sentence sentence, int start, int length,
    int value, boolean isOkay
  ) {
    this.sentence = sentence;
    this.start = start; this.length = length;
    this.value = value; this.isOkay = isOkay;
  }
  
  /** The Sentence in which this Escape was found. */
  public final Sentence sentence;
  
  /** The index of the character in 'sentence' at which this Escape starts. */
  public final int start;
  
  /** The number of characters that make up this Escape. */
  public final int length;
  
  /** The character code represented by this Escape. This is a 32-bit value. The value is not defined unless 'isOkay'. */
  public final int value;
  
  /** 'true' if this Escape represents a syntactically correct escape sequence, otherwise 'false'. */
  public final boolean isOkay;
  
  /** Lexes an Escape in the middle of 'sentence' starting at 'start'. This method always returns an instance of Escape. If the escape sequence is syntactically incorrect, this method returns an Escape representing the correct part, and attaches an error message to 'sentence'.
   * @param sentence the Sentence from which to pluck an Escape.
   * @param start the index of the initial '\' character of the escape sequence.
   * @return a fresh Escape.
   * @throws IllegalArgumentException if 'start' is not the index of a '\' character.
   */
  public static Escape lex(final Sentence sentence, final int start) {
    int i = start;
    if (i>=sentence.length || sentence.get(i++)!='\\') {
      throw new IllegalArgumentException(
        "Escape sequences must start with a \\ character"
      );
    }
    int code = 0;
    while (i<sentence.length) {
      final char c = sentence.get(i);
      if (c>='0' && c<='9') code = (code<<4) + (c-'0');
      else if (c>='A' && c<='F') code = (code<<4) + (c-'A'+10);
      else break;
      i++;
    }
    final int numDigits = i-(start+1);
    boolean isOkay = true;
    if (i<sentence.length && sentence.get(i)=='/') {
      i++;
    } else {
      sentence.addError(
        "This escape sequence does not end correctly. "+
        "There is a / character missing.",
        start, i-start
      );
      isOkay = false;
    }
    if (numDigits<1 || numDigits>8) {
      sentence.addError(
        "This escape sequence must have at least one and at most eight digits "+
        "(it has "+numDigits+").",
        start, i-start
      );
      isOkay = false;
    }
    return new Escape(sentence, start, i-start, code, isOkay);
  }
  
  /* Override things in Object. */
  
  public String toString() {
    return
      "Escape[start="+this.start+
      ", length="+this.length+
      ", value="+this.value+
      ", isOkay="+this.isOkay+
      "]";
  }
  
  /* Test code. */
  
  public static void main(String[] args) {
    final Sentence sentence =
      new Sentence("\\FROG\\4/4\\/**\\4/444\\4".toCharArray());
    System.out.println("sentence = "+sentence.getString(0, sentence.length));
    System.out.println("lex(sentence, 0) = "+lex(sentence, 0));
    System.out.println("lex(sentence, 5) = "+lex(sentence, 5));
    System.out.println("lex(sentence, 9) = "+lex(sentence, 9));
    System.out.println("lex(sentence, 13) = "+lex(sentence, 13));
    System.out.println("lex(sentence, 19) = "+lex(sentence, 19));
    sentence.printErrorReport(System.out, 1000);
  }
}
