package org.sc3d.apt.sss.v3;

/** Represents a lexical analysis of a Sentence. The analysis recognises comments, keywords (including punctuation strings and separator characters), constants, identifers, numbers and strings, which it wraps up as Tokens with appropriate types. All white-space is discarded. */
public class Lex {
  /** Constructs a Lex representing the lexical structure of 'sentence'. Attaches error messages to 'sentence' if appropriate. */
  public Lex(Sentence sentence) {
    this.sentence = sentence;
    final TokenBuffer tb = new TokenBuffer();
    boolean suppress = false; // Prevents multiple 'Illegal character' messages.
    for (int i=0; i<sentence.length; ) {
      Token t = null;
      final char c = sentence.get(i);
      switch (c<CC.length ? CC[c] : 0) {
        case C_WHITESPACE: break;
        case C_LONE: t = new Token(Token.TYPE_WORD, sentence, i, 1); break;
        case C_PUNCTUATION: t = this.lexPunctuation(i); break;
        case C_LETTER: t = this.lexAlphanumeric(i); break;
        case C_DIGIT: t = SSSNumber.lex(sentence, i); break;
        case C_SPECIAL:
          switch(c) {
            case '#': t = this.lexComment(i); break;
            case '"': t = SSSString.lex(sentence, i); break;
            case '\'': t = SSSChar.lex(sentence, i); break;
            default:
              throw new RuntimeException("Don't know what to do with "+c);
          }
          break;
        default:
          if (!suppress) sentence.addError(
            "This is an illegal character, which must not appear except in "+
            "comments and strings",
            i, 1
          );
          suppress = true;
          break;
      }
      if (t!=null) {
        tb.append(t);
        i += t.length;
      } else {
        i++;
      }
    }
    this.tokens = tb.toArray();
    this.numTokens = this.tokens.length;
  }
  
  /* New API. */
  
  /** The Sentence of which this is the lexical analysis. */
  public final Sentence sentence;
  
  /** The number of lexographic Tokens found in 'sentence'. */
  public final int numTokens;
  
  /** Returns one of the lexographic Tokens.
   * @param index the index of the Token, from '0' to 'numTokens-1'.
   */
  public Token getToken(int index) { return this.tokens[index]; }
  
  /* Private. */
  
  /** Reads a maximal string of SSS punctuation characters starting at 'start' and returns a Token of type 'TYPE_WORD'. */
  private Token lexPunctuation(final int start) {
    int i = start;
    while (i<this.sentence.length) {
      final char c = this.sentence.get(i);
      if (c>CC.length || CC[c]!=C_PUNCTUATION) break;
      i++;
    }
    return new Token(Token.TYPE_WORD, this.sentence, start, i-start);
  }
  
  /** Reads a maximal string of alphanumeric characters starting at 'start' and returns a Token of type 'TYPE_WORD', 'TYPE_CONSTANT' or 'TYPE_IDENTIFIER' as appropriate. */
  private Token lexAlphanumeric(final int start) {
    int i = start;
    if (i>=this.sentence.length) throw new IllegalArgumentException();
    char c = this.sentence.get(i++);
    boolean isInitialCapital = c>='A' && c<='Z';
    boolean isAllCapital = isInitialCapital;
    while (i<this.sentence.length) {
      c = this.sentence.get(i);
      if ((c>='a' && c<='z') || (c>='0' && c<='9')) isAllCapital = false;
      else if (c<'A' || c>'Z') break;
      i++;
    }
    final int length = i-start;
    return new Token(
      isAllCapital && length>1 ? Token.TYPE_WORD :
      isInitialCapital ? Token.TYPE_CONSTANT : Token.TYPE_IDENTIFIER,
      this.sentence, start, length
    );
  }
  
  /** Lexes an SSS comment starting at 'start', and returns a Token with type 'TYPE_COMMENT'. */
  private Token lexComment(final int start) {
    int i = start;
    if (i>=this.sentence.length || this.sentence.get(i++)!='#')
      throw new IllegalArgumentException("Comments must start with #");
    while (i<this.sentence.length && this.sentence.get(i)!='\n') i++;
    final Token ans =
      new Token(Token.TYPE_COMMENT, this.sentence, start, i-start);
    if (i>=this.sentence.length) ans.addError(
      "This comment does not end. "+
      "There's a 'new line' character missing at the end of the file."
    );
    return ans;
  }
  
  /** An array of length 'numTokens', containing the lexographic Tokens. */
  private final Token[] tokens;
  
  /** The value stored in 'CC' for white-space characters. */
  private static final int C_WHITESPACE = 1;
  
  /** The value stored in 'CC' for characters that form a Token all on their own. */
  private static final int C_LONE = 2;
  
  /** The value stored in 'CC' for SSS punctuation characters. */
  private static final int C_PUNCTUATION = 3;
  
  /** The vlaue stored in 'CC' for letters. */
  private static final int C_LETTER = 4;
  
  /** The value stored in 'CC' for decimal digits. */
  private static final int C_DIGIT = 5;
  
  /** The value stored in 'CC' for special characters that introduce things. */
  private static final int C_SPECIAL = 6;
  
  /** An array which classifies ASCII characters into one of the 'C_XXX' classes. */
  private static final int[] CC = new int[127];
  static {
    classify(C_WHITESPACE, " \t\n\r");
    classify(C_LONE, ",.;()[]{}");
    classify(C_LETTER, "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz");
    classify(C_DIGIT, "0123456789");
    classify(C_SPECIAL, "#\"'");
    classify(C_PUNCTUATION, "!$%&*+-/:<=>?@\\^_`|~");
  }
  
  private static void classify(int type, String s) {
    for (int i=0; i<s.length(); i++) CC[s.charAt(i)] = type;
  }
  
  /* Test code. */
  
  /** Takes one argument: a filename. Reads the file and lexes it. If there are errors, prints an error report. Otherwise lists the Tokens. */
  public static void main(String[] args) throws java.io.IOException {
    if (args.length!=1) throw new IllegalArgumentException(
      "Syntax: java org.sc3d.apt.sss.v3.Lex <filename>"
    );
    final Sentence sentence = Sentence.readFile(args[0]);
    final Lex me = new Lex(sentence);
    if (sentence.countErrors()>0) {
      sentence.printErrorReport(System.out, 100);
      return;
    }
    for (int i=0; i<me.numTokens; i++)
      System.out.println("tokens["+i+"] = "+me.getToken(i));
  }
}
