package org.sc3d.apt.sss.v3;

/** Represents a bracket matching analysis of a Sentence. */
public class Match {
  /** Constructs a Match. This constructor is private. Use 'make()' instead. */
  private Match(Sentence sentence, Token[] tokens) {
    this.sentence = sentence;
    this.tokens = tokens; this.numTokens = tokens.length;
  }
  
  /** Tries to construct and return a Match for 'lex.sentence'. If the brackets don't match, this method attaches appropriate error messages to the Sentence and returns 'null'. */
  public static Match make(Lex lex) {
    try {
      final TokenBuffer tb = new TokenBuffer();
      final int used = match(tb, lex, 0);
      if (used<lex.numTokens) {
        lex.getToken(used).addError("This bracket is not matched.");
        return null;
      }
      return new Match(lex.sentence, tb.toArray());
    } catch (UnmatchedBracketException e) {
      return null;
    }
  }
  
  /* New API. */
  
  /** The sentence of which this is a bracket matching analysis. */
  public final Sentence sentence;
  
  /** The number of Tokens in 'sentence', after all Tokens inside brackets have been grouped into instances of Bracket. */
  public final int numTokens;
  
  /** Returns the Token with index 'index'.
   * @param index a number from '0' to 'numTokens-1'.
   */
  public Token getToken(int index) { return this.tokens[index]; }
  
  /** Returns all the Tokens in a fresh array of length 'numTokens'. */
  public Token[] toArray() { return (Token[])this.tokens.clone(); }
  
  /* Private. */
  
  /** Thrown by 'match()' for sentences containing unmatched brackets. */
  private static class UnmatchedBracketException extends Exception {
    public UnmatchedBracketException() { super(); }
    public UnmatchedBracketException(String msg) { super(msg); }
  }
  
  /** Reads Tokens from 'lex' starting at Token 'start', groups together Tokens inside brackets into instances of Bracket, appends them to 'tb', and stops before the first enclosing close-bracket or at the end of the Sentence.
   * @return the number of Tokens read.
   * @throws UnmatchedBracketException if the brackets in 'sentence' do not match correctly, and attaches appropriate error messages to 'sentence'.
   */
  private static int match(
    final TokenBuffer tb,
    final Lex lex,
    final int start
  ) throws UnmatchedBracketException {
    for (int i=start; i<lex.numTokens; i++) {
      Token t = lex.getToken(i);
      if (t.length==1) {
        final char c = t.sentence.get(t.start);
        if (c==')' || c==']' || c=='}') return i-start;
        if (c=='(' || c=='[' || c=='{') {
          i++;
          final TokenBuffer contents = new TokenBuffer();
          i += match(contents, lex, i);
          if (i>=lex.numTokens) {
            t.addError("This bracket is not matched.");
            throw new UnmatchedBracketException();
          }
          final Token closing = lex.getToken(i);
          if (!Bracket.matches(t, closing)) {
            t.addError("This bracket matches the wrong kind of bracket.");
            closing.addError("This bracket matches the wrong kind of bracket.");
            throw new UnmatchedBracketException();
          }
          t = new Bracket(lex.sentence, t, contents.toArray(), closing);
        }
      }
      tb.append(t);
    }
    return lex.numTokens-start;
  }
  
  /** The Tokens, after grouping things inside brackets into instances of Bracket. */
  private final Token[] tokens;
  
  /* Test code. */
  
  /** Dumps to 'System.out' the contents of 'tokens' one Token to a line, prefixing each line with 'prefix', and making recursive calls to dump the contents of Brackets. */
  private static void dump(String prefix, Token[] tokens) {
    for (int i=0; i<tokens.length; i++) {
      if (tokens[i] instanceof Bracket) {
        final Bracket b = (Bracket)tokens[i];
        System.out.println(prefix+"["+i+"] = Bracket "+b.opening+b.closing);
        dump(prefix+"["+i+"]", ((Bracket)tokens[i]).toArray());
      } else {
        System.out.println(prefix+"["+i+"] = "+tokens[i]);
      }
    }
  }
  
  /** Takes one argument: a filename. */
  public static void main(String[] args) throws java.io.IOException {
    if (args.length!=1) throw new IllegalArgumentException(
      "Syntax: java org.sc3d.apt.sss.v3.Match <filename>"
    );
    final Sentence sentence = Sentence.readFile(args[0]);
    final Lex lex = new Lex(sentence);
    final Indentation indentation = new Indentation(lex);
    final Match me = make(lex);
    if (sentence.countErrors()!=0) {
      sentence.printErrorReport(System.out, 1000000);
    } else {
      System.out.println("No errors.");
    }
    if (me!=null) dump("", me.toArray());
  }
}
