package org.sc3d.apt.sss.v3;

import java.io.*;
import java.util.*;

/** Represents a parser for a particular Grammar. The algorithm used is a depth-first search using an NDFA. Parsers do not look inside Brackets by default. Subclasses can override the 'postProcess()' method to make Parsers customised for a particular Grammar. */
public class Parser {
  /** Constructs a Parser for 'grammar'. */
  public Parser(Grammar grammar) {
    this.ndfas = new HashMap();
    this.ndfa = this.make(grammar);
  }
  
  /* New API. */

  /** An NDFA that represents the grammar passed to the constructor and used by 'parse()'. */
  public final NDFA ndfa;
  
  /** Parses 'sentence' and returns the result of applying 'postProcess()' to its parse-tree. This method first uses 'sentence' to construct a Lex, a Match and an Indentation. They are constructed in that order so as to make the error messages as helpful as possible even if they collide. If there are any errors so far, this method returns 'null' (it is the caller's responsibility to print an error report). Otherwise, it calls 'parseTokens(match.toArray())', and again returns 'null' if there are any errors. Otherwise, it returns the result of passing the parse-tree to 'postProcess()'. */
  public Object parse(final Sentence sentence) {
    final Lex lex = new Lex(sentence);
    final Match match = Match.make(lex);
    final Indentation indentation = new Indentation(lex);
    if (sentence.countErrors()!=0 || match==null) return null;
    final Tree raw = this.parseTokens(this.ndfa, match.toArray(), null);
    return raw==null ? null : this.postProcess(raw);
  }

  
  /** Parses 'tokens' according to 'ndfa' and returns its parse-tree. If there are any errors, this method attaches appropriate error messages to the Sentence from which the Tokens came and then returns 'null'. If 'tokens.length' is zero, then there is no way of finding the Sentence, so this method does not generate any error messages and simply returns 'null'.
   * @param ndfa an NDFA representing the grammar.
   * @param tokens the Tokens to parse.
   * @param after the first Token after 'tokens', or 'null' if 'tokens' is a whole Sentence. This provides a place to attach an error message in the case that there is something missing at the end of 'tokens'.
   * @return the raw parse-tree.
   */
  public Tree parseTokens(
    NDFA ndfa, final Token[] tokens, Token after
  ) {
    final String afterString =
      after==null ? "end of sentence" : '"'+after.toString()+'"';
    int best = 0;
    TreeSet expected = new TreeSet(); // Contains Strings.
    Stack stack = new Stack(ndfa.initialState, 0, null, null);
    while (stack!=null) {
      final NDFA.State q = stack.q;
      final int pos = stack.pos;
      final PartTree pt = stack.pt;
      stack = stack.next;
      if (q==ndfa.finalState) {
        if (pos==tokens.length) return pt.toTree(); // Yay! Success.
        // Sentence could have ended here.
        if (pos>best) { best = pos; expected = new TreeSet(); }
        if (pos>=best) expected.add(afterString);
      }
      // Loop through the Transitions from 'q' from least to most preferred.
      for (int i=q.count()-1; i>=0; i--) {
        final NDFA.Transition trans = q.get(i);
        if (trans.g==null) {
          if (trans.p==null) { // No-op.
            stack = new Stack(trans.to, pos, pt, stack);
          } else { // Reduce.
            final Tree[] parts = new Tree[trans.p.numParts];
            PartTree newPT = pt;
            for (int j=parts.length-1; j>=0; j--) {
              parts[j] = newPT.toTree();
              newPT = newPT.prev;
            }
            final Tree.Production prod =
              new Tree.Production(trans.p.name, parts);
            final PartTree.NonTerminal pnt = (PartTree.NonTerminal)newPT;
            newPT = new PartTree.NonTerminal(
              pnt.prev,
              new PartTree.Production(pnt.repeats, prod)
            );
            stack = new Stack(trans.to, pos, newPT, stack);
          }
        } else {
          if (trans.g.isTerminal) { // Shift.
            final Grammar.Terminal g = (Grammar.Terminal)trans.g;
            if (pos<tokens.length && g.accepts(tokens[pos])) {
              // Found the right kind of Token.
              final PartTree newPT =
                new PartTree.Terminal(pt, new Terminal(tokens[pos], trans));
              stack = new Stack(trans.to, pos+1, newPT, stack);
            } else {
              // Found the wrong kind of Token.
              if (pos>best) { best = pos; expected = new TreeSet(); }
              if (pos>=best) expected.add(g.toString());
            }
          } else { // Open.
            final PartTree newPT = new PartTree.NonTerminal(pt, null);
            stack = new Stack(trans.to, pos, newPT, stack);
          }
        }
      }
    }
    // There was no path through the NDFA. Construct an error message.
    final StringBuffer sb = new StringBuffer("Expected: ");
    final Iterator it = expected.iterator();
    if (it.hasNext()) {
      sb.append(it.next());
      while (it.hasNext()) sb.append(" or ").append(it.next());
    }
    sb.append(", but not ");
    if (best<tokens.length) {
      sb.append("\"").append(tokens[best]).append("\".");
      tokens[best].addError(sb.toString());
    } else {
      sb.append(afterString).append(".");
      if (after!=null) {
        after.addError(sb.toString());
      } else if (best>0) {
        tokens[best-1].sentence.addErrorAtEnd(sb.toString());
      } else {
        // No way to find the Sentence, so do nothing.
      }
    }
    return null;
  }
  
  /** Subclasses which are specialised for a paticular Grammar can override this method in order to post-process the parse-tree. In particular, they might want to call the parse() methods of Brackets. If it finds any errors, this method must attach appropriate error messages to the Sentence from which the Tokens came and then return 'null'. The default implementation simply returns 'raw'.
   * @param raw the non-null parse-tree.
   * @return a grammar-specific data structure, or 'null' if there were errors.
   */
  public Object postProcess(Tree raw) { return raw; }
  
  /** Prompts for input on 'System.in', parses it, and prints the parse-tree. If you write a subclass of Parser, you may want to provide a 'main()' method that tests the parser by calling this method.
   * <p>In more detail, this method reads a line at a time until it is given a blank line. It concatenates the non-blank lines into a Sentence, parses it, and prints out the value returned by 'postProcess()'. If 'postProcess()' is not overridden, then the default implementation does not parse inside brackets. The method loops until it finds an "end of file" or encounters an IOException.
   */
  public void interactiveTest() {
    try {
      final BufferedReader in =
        new BufferedReader(new InputStreamReader(System.in));
      while (true) {
        // Read lines until the next blank line.
        final StringBuffer sb = new StringBuffer();
        String prompt = "\n>";
        while (true) {
          System.out.print(prompt);
          final String line = in.readLine();
          if (line==null) { System.out.println("Exit."); return; }
          if ("".equals(line)) break;
          sb.append(line).append('\n');
          prompt = ".";
        }
        final Sentence sentence = new Sentence(sb.toString().toCharArray());
        final Object tree = this.parse(sentence);
        if (tree==null) {
          sentence.printErrorReport(System.out, 10);
        } else {
          System.out.println(tree);
        }
      }
    } catch (IOException e) { e.printStackTrace(); }
  }
  
  /* Private. */
  
  /** Returns the NDFA for 'grammar', constructing a new one if necessary. */
  private NDFA make(Grammar grammar) {
    NDFA ans = (NDFA)this.ndfas.get(grammar);
    if (ans==null) {
      ans = NDFA.make(new NDFA.State(), new NDFA.State(), grammar);
      this.ndfas.put(grammar, ans);
    }
    return ans;
  }
  
  /** A Map from Grammars to NDFAs. This is used to ensure that Grammars are converted to NDFAs at most once each. */
  private HashMap ndfas;
  
  ////////////////////////////////////////////////////////////////////////////
  
  /** Represents a stack of back-tracking alternatives. */
  private static class Stack {
    /** Constructs a Stack given valeus for its fields. */
    public Stack(NDFA.State q, int pos, PartTree pt, Stack next) {
      this.q = q; this.pos = pos; this.pt = pt; this.next = next;
    }
    
    /** The most preferred NDFA.State. */
    public final NDFA.State q;
    
    /** The index of the next Token to parse for 'q'. */
    public final int pos;
    
    /** The parse-tree so far for 'q'. */
    public final PartTree pt;
    
    /** All the less preferred alternatives. */
    public final Stack next;
  }
  
  ////////////////////////////////////////////////////////////////////////////
  
  /** A subclass of Tree.Terminal whose 'parse()' method calls 'parseTokens()'. */
  private class Terminal extends Tree.Terminal {
    /** Constructs a Terminal given values for its fields. */
    public Terminal(Token t, NDFA.Transition trans) {
      super(t); this.trans = trans;
    }
    
    /** The NDFA.Transition that caused this Terminal to be constructed. */
    public final NDFA.Transition trans;
    
    /** Parses the contents of 't' according to 'trans.ndfa'. If 'trans.ndfa' is null, it is first calculated by passing 'trans.g.getGrammarOfContents()' to 'make()'. Do not call this method unless 't' is a bracket. */
    public Tree parse() {
      if (this.trans.ndfa==null) {
        final Grammar.Terminal g = (Grammar.Terminal)trans.g;
        this.trans.ndfa = Parser.this.make(g.getGrammarOfContents());
      }
      final Bracket b = (Bracket)this.t;
      return Parser.this.parseTokens(this.trans.ndfa, b.toArray(), b.closing);
    }
  }
}
