package org.sc3d.apt.sss.v3;

/** Represents a half-finished Tree. The Tree is constructed left-to-right, and Productions are used to join things together as soon as is possible. Unless your requirements are unusual, you probably don't need to use PartTrees directly, as Parser wraps it all up for you.
 * <p>The structure of a PartTree closely mimics that of Tree. The main difference is that a PartTree may not yet be a single tree, because we may not have all the branches needed to construct the root Tree.NonTerminal. Instead it consists of a linked list of zero or more trees.
 * <p>Instances are immutable, and can therefore be shared wherever it is convenient. 'null' represents the empty list.
 */
abstract class PartTree {
  /** Don't want any instances except of the inner classes. */
  public PartTree(PartTree prev, boolean isTerminal) {
    this.prev = prev; this.isTerminal = isTerminal;
  }
  
  /* New API. */
  
  /** A PartTree representing all but the last tree of this PartTree, or 'null' if this PartTree contains only one tree. */
  public final PartTree prev;
  
  /** 'true' if the last tree of this PartTree is a single Token matching a Grammar.Terminal, or 'false' if the last tree matches a Grammar.NonTerminal. */
  public final boolean isTerminal;
  
  /** Converts the last tree of this PartTree into a finished Tree. */
  public abstract Tree toTree();
  
  /** For debugging. */
  public abstract String toString();
  
  ////////////////////////////////////////////////////////////////////////////

  /** The subclass of PartTree with 'isTerminal', which represents a PartTree whose last tree is a Tree.Terminal. */
  static class Terminal extends PartTree {
    /** Constructs a Terminal given values for its fields. */
    public Terminal(PartTree prev, Tree.Terminal t) {
      super(prev, true);
      this.t = t;
    }
    
    /** The Token that the parser found to satisfy the Grammar.Terminal. */
    public final Tree.Terminal t;
    
    /** Returns a Tree.Terminal. */
    public Tree toTree() { return this.t; }
    
    public String toString() {
      return
        "PartTree.Terminal[t="+this.t+
        ", prev="+this.prev+"]";
    }
  }
  
  ////////////////////////////////////////////////////////////////////////////

  /** The subclass of PartTree with '!isTerminal', which represents a PartTree whose last tree matches a Grammar.NonTerminal. */
  static class NonTerminal extends PartTree {
    /** Constructs a NonTerminal given values for its fields. */
    public NonTerminal(PartTree prev, Production repeats) {
      super(prev, false);
      this.repeats = repeats;
    }
    
    /** A Production representing the trees that the parser has found so far to satisfy the Grammar.NonTerminal. If it is not optional, there will eventually be at least one element. If it is not repeatable, there will always be at most one element. */
    public final Production repeats;
    
    /** Returns a Tree.NonTerminal. */
    public Tree toTree() {
      return new Tree.NonTerminal(Production.toArray(this.repeats));
    };
    
    public String toString() {
      return
        "PartTree.NonTerminal[repeats="+repeats+
        ", prev="+prev+"]";
    }
  }
  
  ////////////////////////////////////////////////////////////////////////////

  /** Represents a linked list of Tree.Productions. Instances are immutable. 'null' represents the empty list. */
  static class Production {
    /** Constructs a Production given values for its fields. */
    public Production(Production prev, Tree.Production last) {
      this.prev = prev; this.last = last;
    }
    
    /** A Production representing all but the last element of this list, or 'null' if there is only one element. */
    public final Production prev;
    
    /** The last element of this list. */
    public final Tree.Production last;
    
    /** Returns the elements of 'p' in a fresh array. This method is static in order to cope with the case 'p==null'. */
    public static Tree.Production[] toArray(final Production p) {
      int i = 0;
      for (Production q = p; q!=null; q = q.prev) i++;
      final Tree.Production[] ans = new Tree.Production[i];
      for (Production q = p; q!=null; q = q.prev) ans[--i] = q.last;
      return ans;
    }
    
    /** For debugging. */
    public String toString() {
      return "PartTree.Production[last="+this.last+", prev="+this.prev+"]";
    }
  }
}
