package org.sc3d.apt.sss.v3;

/** A subclass of Token that represents an SSS number, or any prefix thereof. */
public class SSSNumber extends Token {
  /** Constructs an SSSNumber, given values for its fields. It's type will be 'TYPE_NUMBER'. */
  public SSSNumber(
    Sentence sentence,
    int start,
    int integerLength,
    int fractionLength,
    int exponentLength
  ) {
    super(
      TYPE_NUMBER,
      sentence,
      start,
      integerLength+fractionLength+exponentLength
    );
    this.integerLength = integerLength;
    this.fractionLength = fractionLength;
    this.exponentLength = exponentLength;
  }

  /** Constructs an SSSNumber by reading from the middle of a Sentence. This method always returns an SSSNumber. If there is a syntax error, the returned SSSNumber will represent the syntactically correct part, and an error will be added to 'sentence'.
   * @param sentence the sentence to read.
   * @param start the index of the first character of the number.
   * @return a new SSSNumber.
   */
  public static SSSNumber lex(final Sentence sentence, final int start) {
    int i = start;
    // Lex the integer part.
    if (i>=sentence.length || !isDigit(sentence.get(i))) {
      sentence.addError(
        "There's a digit missing here. A number must start with a decimal "+
        "digit, even if it is written in hexadecimal.",
        i, 1
      );
    } else {
      i++;
      while (i<sentence.length && isHexDigit(sentence.get(i))) i++;
    }
    // Lex the fraction part.
    final int fraction = i;
    if (i<sentence.length && sentence.get(i)=='.') {
      i++;
      while (i<sentence.length && isHexDigit(sentence.get(i))) i++;
    }
    // Lex the exponent part.
    final int exponent = i;
    if (i<sentence.length && isBase(sentence.get(i))) {
      i++;
      if (i<sentence.length && sentence.get(i)=='-') i++;
      final int firstDigit = i;
      while (i<sentence.length && isDigit(sentence.get(i))) i++;
      if (i<=firstDigit) sentence.addError(
        "There's a digit missing here. If you write a number with an explicit "+
        "base character, you must also include an exponent. The exponent must "+
        "be written in decimal. Perhaps you meant '0'?",
        i, 1
      );
    }
    // Construct the Token.
    final SSSNumber ans = new SSSNumber(
      sentence,
      start,
      fraction-start,
      exponent-fraction,
      i-exponent
    );
    // Check all the digits are valid for the base.
    char max;
    switch (ans.getBase()) {
      case 'b': max = '1'; break;
      case 'q': max = '3'; break;
      case 'o': max = '7'; break;
      case 'd': max = '9'; break;
      case 'h': max = 'F'; break;
      default: throw new RuntimeException("Eh?");
    }
    final String msg =
      "This digit is too big for this number base. It can be at most "+max+".";
    for (i=start; i<fraction; i++) {
      if (sentence.get(i)>max) sentence.addError(msg, i, 1);
    }
    for (i=fraction+1; i<exponent; i++) {
      if (sentence.get(i)>max) sentence.addError(msg, i, 1);
    }
    // Return.
    return ans;
  }
  
  /* New API. */
  
  /** The number of characters occupied by the integer part of this SSSNumber. The integer part (if syntactically correct) consists of one or more hexadecimal digits, of which the first must be a decimal digit. */
  public final int integerLength;
  
  /** The number of characters occupied by the optional fractional part of this SSSNumber, if present, or '0', if absent. If the fractional part is present, it consists of a '.' followed by zero or more hexadecimal digits. An absent fractional part means the same as "." or ".0". */
  public final int fractionLength;
  
  /** The number of characters occupied by the optional exponent part of this SSSNumber, if present, or '0', if absent. If the exponent part is present, it consists of a single character identifying the base in which this number is written, an optional '-' sign, and one or more decimal digits. The base character can be one of 'b' for binary, 'q' for quarternary, 'o' for octal, 'd' for decimal, or 'h' for hexadecimal. An absent exponent means the same as an exponent of "d0". */
  public final int exponentLength;

  /** Returns the digits of the integer part of this SSSNumber. */
  public String getIntegerDigits() {
    return this.sentence.getString(this.start, this.integerLength);
  }
  
  /** Returns the digits of the fractional part of this SSSNumber. If the fractional part is absent or ".", this method returns "0". */
  public String getFractionDigits() {
    if (this.fractionLength<2) return "0";
    final int fraction = this.start+this.integerLength;
    return this.sentence.getString(fraction+1, this.fractionLength-1);
  }
  
  /** Returns the base character of this SSSNumber. If the exponent part is absent, this method returns 'd'. See also 'getRadix()'. */
  public char getBase() {
    if (this.exponentLength<1) return 'd';
    final int exponent = this.start+this.integerLength+this.fractionLength;
    return this.sentence.get(exponent);
  }
  
  /** Returns 'true' if the exponent part is present and includes the optional '-' sign. */
  public boolean isExponentNegative() {
    if (this.exponentLength<2) return false;
    final int exponent = this.start+this.integerLength+this.fractionLength;
    return this.sentence.get(exponent+1)=='-';
  }
  
  /** Returns the digits of the exponent part of this SSSNumber. If the exponent is absent or (erroneously) has no digits, this method returns "0". */
  public String getExponentDigits() {
    if (this.exponentLength<2) return "0";
    final int exponent = this.start+this.integerLength+this.fractionLength;
    int skip = this.sentence.get(exponent+1)=='-' ? 2 : 1;
    if (this.exponentLength<=skip) return "0";
    return this.sentence.getString(exponent+skip, this.exponentLength-skip);
  }
  
  /** Returns the concatenation of the integer and fraction digits. This can be combined with 'getRadix()' and 'getShift()' to evaluate the number. */
  public String getMantissa() {
    String ans = this.getIntegerDigits();
    if (this.fractionLength>1) ans += this.getFractionDigits();
    return ans;
  }
  
  /** Returns '2', '4', '8', '10' or '16' depending on the base in which this number is written. See also 'getBase()'. */
  public int getRadix() {
    switch (this.getBase()) {
      case 'b': return 2;
      case 'q': return 4;
      case 'o': return 8;
      case 'd': return 10;
      case 'h': return 16;
      default: throw new RuntimeException("Eh?!");
    }
  }
  
  /** Returns the exponent minus the number of fractional digits. In other words, this method returns the number of zeros that must be appended to 'getMantissa()' to evaluate this SSSNumber. This method can return a negative number. */
  public int getShift() {
    int ans = Integer.parseInt(this.getExponentDigits());
    if (this.isExponentNegative()) ans = -ans;
    if (this.fractionLength>1) ans -= this.fractionLength-1;
    return ans;
  }
  
  /* Private. */
  
  private static boolean isDigit(char c) {
    return c>='0' && c<='9';
  }
  
  private static boolean isHexDigit(char c) {
    return (c>='0' && c<='9') || (c>='A' && c<='F');
  }
  
  private static boolean isBase(char c) {
    return c=='b' || c=='q' || c=='o' || c=='d' || c=='h';
  }
}
