package org.sc3d.apt.sss.v3;

/** A subclass of Token that represents a literal String. The 'type' field of SSSStrings is always 'Token.TYPE_STRING'. This class adds to Token a field 'value' which returns the value of the literal string when all escape sequences have been replaced by characters. */
public class SSSString extends Token {
  /** Constructs an SSSString, given values for its fields. */
  public SSSString(Sentence sentence, int start, int length, String value) {
    super(Token.TYPE_STRING, sentence, start, length);
    this.value = value;
  }
  
  /** Constructs an SSSString by reading from the middle of a Sentence. This method always returns a non-null SSSString. If there is a syntax error, this method returns an SSSString which represents the syntactically correct part, and attaches an appropriate error message to the Sentence. You can detect errors by checking whether the 'value' field is null.
   * @param sentence the sentence from which to pluck a literal string.
   * @param start the index of the opening double quote character of the literal string.
   * @throws IllegalArgumentException if 'sentence.get(start)' is not a double quote character.
   * @return a new SSSString.
   */
  public static SSSString lex(final Sentence sentence, final int start) {
    int i = start;
    if (i>=sentence.length || sentence.get(i)!='"') {
      throw new IllegalArgumentException(
        "Literal strings must start with a double quote character"
      );
    }
    i++;
    final StringBuffer value = new StringBuffer();
    boolean isOkay = true;
    while (i<sentence.length) {
      final char c = sentence.get(i);
      if (c=='"') { // End of string.
        return new SSSString(
          sentence, start, i+1-start,
          isOkay ? value.toString() : null
        );
      } else if (c=='\\') { // Start of escape sequence.
        final Escape e = Escape.lex(sentence, i);
        if (e.isOkay) value.append((char)e.value);
        else isOkay = false;
        i += e.length;
      } else { // Ordinary character in string.
        value.append(c);
        i++;
      }
    }
    // String did not end.
    sentence.addError(
      "This string does not end. There is a \" character missing somewhere.",
      start, i-start
    );
    return new SSSString(sentence, start, i-start, null);
  }
  
  /* New API. */
  
  /** The value of this SSSString, or 'null' if the string literal was not syntactically correct. Note that characters in this String are truncated to the 16-bit code used by Java. */
  public final String value;
  
  /* Test code. */
  
  public static void main(String[] args) {
    final String s = "BEFORE \"\\22/Hello, World!\\22/\" AFTER";
    final Sentence sentence = new Sentence(s.toCharArray());
    final SSSString me = lex(sentence, 7);
    sentence.printErrorReport(System.out, 1000000);
    System.out.println("me = '"+me+"'");
    System.out.println("me.value = '"+me.value+"'");
  }
}
